import numpy as np
from inspect import signature


# startVector must be a VECTOR!!!
# error indicates how close to zero must be the derrivative
# sd = steepest descent ; sam = simulated annealing method ; ne = Nelder Mead method
def fmin(func, startVector, error=0.000001, steps=1000, stepSize=0.001, method='ne'):
    sig = signature(func)
    argNum = len(sig.parameters)
    startVector = np.array(startVector)
    startVector = np.ndarray.flatten(startVector)
    currentVector = startVector
    currentVshape = np.shape(currentVector)
    if currentVshape[0] != argNum:
        raise Exception('Number of function arguments doesn\'t match number of starting point parameters')
    else:
        if method == 'sd':
            # calculate gradient
            def grad(currx, h):
                dy = np.array([])
                for i in range(argNum):
                    currx[i] += h
                    f1 = func(*currx)
                    currx[i] -= 2 * h
                    f2 = func(*currx)
                    tmpdy = (f1 - f2) / (2 * h)
                    if i == 0:
                        dy = tmpdy
                    else:
                        dy = np.r_[dy, tmpdy]
                return dy

            i = 1
            currentVector = startVector
            grdmag = 1
            while i < steps and grdmag > error:
                grd = grad(currentVector, stepSize)
                grdmag = np.sqrt(np.array([grd]).dot(grd))
                currentVector = currentVector - 0.9 * grd
                i += 1
        #            print(f'steepest descent steps: {i}')
        elif method == 'sam':  # simulated annealing !!! doesn't work
            tmin = error
            tmax = 1000
            t = tmax
            i = 0
            bestVector = startVector
            currentVector = bestVector
            while t > tmin:
                x = np.random.random(argNum)
                randomVector = x
                dE = func(*randomVector) - func(*currentVector)
                prob = np.exp(-dE / t)
                rand = np.random.choice(a=[1, 0], size=1, p=[prob, 1 - prob])
                if rand == 1:
                    currentVector = randomVector
                else:
                    if func(*currentVector) < func(*bestVector):
                        bestVector = currentVector
                i += 1
                t = tmax / (i + 1)
            currentVector = bestVector
        elif method == 'ne':  # Nelder Mead
            bestP = startVector
            if argNum > 0:
                p = np.array([bestP])
                for h in range(argNum):
                    pnew = np.random.random(argNum)
                    p = np.vstack([p, pnew])
                alpha = 1
                gamma = 2
                rho = 0.5
                sigma = 0.5
                #               j = -1
                fpref = 1
                fpexp = 1
                fpcont = 1
                diff = 1
                while diff > error:
                    #                    j += 1
                    fvec = np.array([])
                    for h in range(argNum + 1):
                        fvec = np.append(fvec, func(*p[h]))
                    pworst = np.argmax(fvec)
                    pbest = np.argmin(fvec)
                    diff = np.abs((fvec[pworst] - fvec[pbest]) / (fvec[pbest] + 1e-9))
                    tmp = np.delete(p, pworst, 0)
                    centroid = np.sum(tmp, axis=0) / np.shape(tmp)[0]
                    bestP = centroid
                    # reflection
                    npref = centroid + alpha * (centroid - p[pworst])
                    fpref = func(*npref)
                    if fvec[pbest] <= fpref < fvec[pworst]:
                        p[pworst] = npref
                        continue
                    elif fpref < fvec[pbest]:
                        # expansion
                        npexp = npref + gamma * np.abs(npref - centroid)
                        fpexp = func(*npexp)
                        if fpexp < fpref:
                            p[pworst] = npexp
                            continue
                        else:
                            p[pworst] = npref
                            continue
                    else:
                        # contraction
                        if fpref < fvec[pworst]:
                            npcont = centroid + rho * (npref - centroid)
                        elif fpref >= fvec[pworst]:
                            npcont = centroid - rho * (centroid - p[pworst])
                        fpcont = func(*npcont)
                        if fpcont < fvec[pworst]:
                            p[pworst] = npcont
                            continue
                        # shrink
                        for k in range(np.shape(p)[0]):
                            if k == pbest:
                                continue
                            else:
                                p[k] = p[pbest] + sigma * (p[k] - p[pbest])
                currentVector = bestP
    #                print(f'ne steps: {j}')
    return currentVector



from OpenGL.GLUT import *
from OpenGL.GL import *
from OpenGL.GLU import *
import numpy as np
import sys

_x = np.array([[]])
_y = np.array([[]])
_xmin = _xmax = _ymin = _ymax = 0
_xlabel = ""
_ylabel = ""
_title = ""
_fastMode = False
_sampleRate = 0
_multplot = False


def __init():
    glClearColor(1.0, 1.0, 1.0, 1.0)
    global _xmin
    global _xmax
    global _ymin
    global _ymax

    gluOrtho2D(_xmin, _xmax, 2 * _ymin, 2 * _ymax)


def __plotting():
    global _title
    glutInit(sys.argv)
    glutInitDisplayMode(GLUT_SINGLE | GLUT_RGB)
    glutInitWindowSize(400, 400)
    glutInitWindowPosition(400, 400)
    glutCreateWindow(_title)
    glutDisplayFunc(__plotFunc)
    __init()
    glutMainLoop()


def __plotFunc():
    global _x
    global _y
    global _xmin
    global _xmax
    global _ymin
    global _ymax
    global _fastMode
    global _sampleRate
    global _multplot
    glClear(GL_COLOR_BUFFER_BIT)
    glColor3f(0.0, 0.0, 0.0)
    glLineWidth(2.0)
    glPointSize(2.0)

    # making plot axis
    glBegin(GL_LINES)
    glVertex2f(_xmin, 0.0)  # x axis
    glVertex2f(_xmax, 0.0)
    glVertex2f(0.0, 2 * _ymax)  # y axis
    glVertex2f(0.0, 2 * _ymin)

    glVertex2f(0.0, 2 * _ymax)  # y axis arrow
    glVertex2f(0.0 - _xmax / 40, 2 * (_ymax - _ymax / 30))
    glVertex2f(0.0, 2 * _ymax)
    glVertex2f(0.0 + _xmax / 40, 2 * (_ymax - _ymax / 30))

    glVertex2f(_xmax, 0.0)  # x axis arrow
    glVertex2f(_xmax - _xmax / 30, 0.0 + _ymax / 30)
    glVertex2f(_xmax, 0.0)
    glVertex2f(_xmax - _xmax / 30, 0.0 - _ymax / 30)

    glEnd()
    glFlush()

    clr = 0
    if _fastMode:
        if _multplot:
            rxv, cxv = np.shape(_x)
            for i in range(rxv):
                if clr == 0:
                    glColor3f(0.0, 0.0, 1.0)  # blue
                elif clr == 1:
                    glColor3f(0.0, 1.0, 0.0)  # green
                elif clr == 2:
                    glColor3f(1.0, 0.0, 0.0)  # red
                elif clr == 3:
                    glColor3f(0.0, 1.0, 1.0)  # cyan
                elif clr == 4:
                    glColor3f(1.0, 0.0, 1.0)  # pink
                elif clr == 5:
                    glColor3f(1.0, 1.0, 0.0)  # yellow
                j = 0
                while j < np.size(_x[i]):
                    glBegin(GL_POINTS)
                    glVertex2f(_x[i][j], _y[i][j])
                    glEnd()
                    glFlush()
                    j += _sampleRate  # skip size
                clr += 1
        else:
            glColor3f(0.0, 0.0, 1.0)
            j = 0
            while j < np.size(_x):
                glBegin(GL_POINTS)
                glVertex2f(_x[j], _y[j])
                glEnd()
                glFlush()
                j += _sampleRate  # skip size
    else:
        if _multplot:
            rxv, cxv = np.shape(_x)
            for i in range(rxv):
                if clr == 0:
                    glColor3f(0.0, 0.0, 1.0)  # blue
                elif clr == 1:
                    glColor3f(0.0, 1.0, 0.0)  # green
                elif clr == 2:
                    glColor3f(1.0, 0.0, 0.0)  # red
                elif clr == 3:
                    glColor3f(0.0, 1.0, 1.0)  # cyan
                elif clr == 4:
                    glColor3f(1.0, 0.0, 1.0)  # pink
                elif clr == 5:
                    glColor3f(1.0, 1.0, 0.0)  # yellow
                for j in range(np.size(_x[i]) - 1):
                    glBegin(GL_LINES)
                    glVertex2f(_x[i][j], _y[i][j])
                    glVertex2f(_x[i][j + 1], _y[i][j + 1])
                    glEnd()
                    glFlush()
                clr += 1
        else:
            glColor3f(0.0, 0.0, 1.0)
            for j in range(np.size(_x) - 1):
                glBegin(GL_LINES)
                glVertex2f(_x[j], _y[j])
                glVertex2f(_x[j + 1], _y[j + 1])
                glEnd()
                glFlush()


def plot(x, y, xlabel="x", ylabel="y", title="plot of y(x)", fastMode=False, sampleRate=0):
    global _x
    global _y
    global _xmin
    global _xmax
    global _ymin
    global _ymax
    global _xlabel
    global _ylabel
    global _title
    global _fastMode
    global _sampleRate
    if np.size(x) != np.size(y):
        raise Exception(f'x and y vectors must be the same size. Length of x={np.size(x)}, y={np.size(y)}')
    if sampleRate == 0 and fastMode:
        raise Exception("you've chosen fastMode so you need to specify sampling rate with sampleRate argument")
    else:
        _sampleRate = sampleRate
    _x = np.copy(x)
    _y = np.copy(y)
    _xmin = np.amin(_x)
    _xmax = np.amax(_x)
    _ymin = np.amin(_y)
    _ymax = np.amax(_y)
    _xlabel = xlabel
    _ylabel = ylabel
    _title = title
    _fastMode = fastMode

    __plotting()


def plotMulti(xvect, yvect, labelvect=None, title="multiple plots", fastMode=False, sampleRate=0):
    # TODO  labels and legends
    if labelvect is None:
        labelvect = []
    global _x
    global _y
    global _xmin
    global _xmax
    global _ymin
    global _ymax
    global _title
    global _multplot
    global _fastMode
    global _sampleRate
    rxv, cxv = np.shape(xvect)
    ryv, cyv = np.shape(yvect)
    tempvect = np.copy(xvect)
    if rxv < ryv:
        if rxv == 1:
            for i in range(ryv - 1):
                xvect = np.row_stack((xvect, tempvect))
    print(xvect)
    if sampleRate == 0 and fastMode:
        raise Exception("you've chosen fastMode so you need to specify sampling rate with sampleRate argument")
    else:
        _sampleRate = sampleRate
    _fastMode = fastMode
    _multplot = True
    lblvsize = np.size(labelvect)
    if lblvsize < ryv:
        for i in range(ryv - lblvsize):
            labelvect = np.append(labelvect, f'y{i}')
    _title = title
    _x = np.copy(xvect)
    _y = np.copy(yvect)
    for xv in _x, _y:
        xmintemp = np.amin(xv)
        xmaxtemp = np.amax(xv)
        if xmintemp < _xmin:
            _xmin = xmintemp
        if xmaxtemp > _xmax:
            _xmax = xmaxtemp
    for yv in _y:
        ymintemp = np.amin(yv)
        ymaxtemp = np.amax(yv)
        if ymintemp < _ymin:
            _ymin = ymintemp
        if ymaxtemp < _ymax:
            _ymax = ymaxtemp
    if (np.abs(_ymax) - np.abs(_ymin)) < 0:
        _ymax = np.abs(_ymin)
    else:
        _ymin = - np.abs(_ymax)

    if (np.abs(_xmax) - np.abs(_xmin)) < 0:
        _xmax = np.abs(_xmin)
    else:
        _xmin = - np.abs(_xmax)

    __plotting()



import models as m
import time_response as tr

s = m.tf([1, 2.5], [1, 2, 1.18])

sssys = m.tf2ss(s)
# print(sssys)

A = [[-4, 0],
     [1.5, -2.1]]
B = [1, 5]
C = [3.2, 1]
D = [0]

sys = m.ss(A,B,C,D)
m.phasePortrait(sys ,t=10)




import sympy as sp
import matplotlib.pyplot as plt
import numpy as np

s = sp.symbols('s', real=False)
t,w = sp.symbols('t w', real=True)
j = sp.I

def tf(numCoef=0,denCoef=1):    #order is: coef of highest power of s->lowest
    num = 0
    den = 0
    numList = numCoef[::-1]     #small arrays, no need for numpy quickness
    denList = denCoef[::-1]
    if(numCoef!=0 and denCoef!=0): 
        for i,coef in enumerate(numList):   #for this to work we need reversed order
            num += coef*(s**i)
            for i,coef in enumerate(denList):
                den += coef*(s**i)
    return num/den
     

def laplace(func):
    sfunc = sp.integrate(func*sp.exp(-s*t),(t,0,sp.oo))
    return sfunc

def invlaplace(sys):
    tfunc = sp.inverse_laplace_transform(sys,s,t)
    return tfunc

def step(sys,time=1):
    stepS = 1/s
    dt = time/1000
    timeNow = 0
    responseVec = np.array([])
    timeVec = np.array([])
    Sresponse = stepS*sys
    response = invlaplace(Sresponse)
    while(timeNow<time):
        temp = response.evalf(subs={t:timeNow})
        # print(temp)
        # print(timeNow)
        if(timeNow>0):
            responseVec = np.append(responseVec,temp)
            timeVec = np.append(timeVec,timeNow)
        timeNow+=dt
    fig = plt.subplot()
    fig.plot(timeVec, responseVec)
    fig.set_xlabel('Time [s]')
    fig.set_ylabel('Amplitude')
    fig.set_title("Step response")
    return fig
        

        



import numpy as np
import models as md


def __setupArrs(system, U, initX=0, time=10, h=0.001, to=0):
    A = B = C = D = np.array([])
    if isinstance(system, md.tf):
        systemSS = md.tf2ss(system)
        A = systemSS.A
        B = systemSS.B
        C = systemSS.C
        D = systemSS.D
    else:
        A = system.A
        B = system.B
        C = system.C
        D = system.D
    rowNum = np.shape(A)[0]
    if isinstance(initX, np.ndarray):
        Xo = np.array([initX])
    elif initX == 0:
        Xo = np.zeros((1, rowNum))
    else:
        temp = np.full((rowNum, 1), initX)
        Xo = temp
    X = np.transpose(Xo)
    Y = np.array([])
    C = np.array([C])
    Yo = np.dot(C, X) + D
    Y = np.append(Y, Yo)
    T = np.array([])
    T = np.append(T, to)
    t = to
    N = int((time - to) / h)  # number of simulation steps
    return A, B, C, D, X, Y, T, N, t


# calculating time domain response and state trajectory of LTI system using Explicit (Forward) Euler method [time in
# seconds]
def solveEE(system, U, initX=0, time=10, h=0.001, to=0):
    A, B, C, D, X, Y, T, N, t = __setupArrs(system, U, initX, time, h, to)
    Udlt = False
    if U == 'delta':
        U = 1
        Udlt = True
    for i in range(N):
        t += h
        Xprev = X[:, [i]]  # ith column of X
        Xnext = Xprev + h * (np.dot(A, Xprev) + np.dot(B, U))  # Forward Euler formula
        X = np.c_[X, Xnext]  # adding new column
        Ytemp = np.dot(C, Xnext)
        Y = np.append(Y, Ytemp)
        T = np.append(T, t)
        if Udlt:
            U = 0
            Udlt = False
    return Y, T, X


# calculating time domain response and state trajectory of LTI system using Implicit (Backward) Euler method
def solveIE(system, U, initX=0, time=10, h=0.001, to=0):
    A, B, C, D, X, Y, T, N, t = __setupArrs(system, U, initX, time, h, to)
    Udlt = False
    if U == 'delta':
        U = 1
        Udlt = True
    for i in range(N):
        t += h
        Xprev = X[:, [i]]
        XFE = Xprev + h * (np.dot(A, Xprev) + np.dot(B, U))  # takes Implicit Euler method's step for approximation
        Xnext = Xprev + h * (np.dot(A, XFE) + np.dot(B, U))
        X = np.c_[X, Xnext]
        Ytemp = np.dot(C, Xnext)
        Y = np.append(Y, Ytemp)
        T = np.append(T, t)
        if Udlt:
            U = 0
            Udlt = False
    return Y, T, X


# calculating time domain response and state trajectory of LTI system using Trapezoidal method
def solveTrap(system, U, initX=0, time=10, h=0.001, to=0):
    A, B, C, D, X, Y, T, N, t = __setupArrs(system, U, initX, time, h, to)
    Udlt = False
    if U == 'delta':
        U = 1
        Udlt = True
    for i in range(N):
        t += h
        Xprev = X[:, [i]]
        Xnprim = np.dot(A, Xprev) + np.dot(B, U)
        XFE = Xprev + h * Xnprim
        Xnext = Xprev + (h / 2) * (
                    Xnprim + (np.dot(A, XFE) + np.dot(B, U)))  # takes average of the slopes from IE and EE methods
        X = np.c_[X, Xnext]
        Ytemp = np.dot(C, Xnext)
        Y = np.append(Y, Ytemp)
        T = np.append(T, t)
        if Udlt:
            U = 0
            Udlt = False
    return Y, T, X


# calculating time domain response and state trajectory of LTI system using Runge-Kutta 4 method
# X' = f(X) => X' = A*X + B*U
def solveRK4(system, U, initX=0, time=10, h=0.001, to=0):
    A, B, C, D, X, Y, T, N, t = __setupArrs(system, U, initX, time, h, to)
    k1 = k2 = k3 = k4 = np.array([])
    Udlt = False
    if U == 'delta':
        U = 1
        Udlt = True
    for i in range(N):
        t += h
        Xprev = X[:, [i]]
        f = np.dot(A, Xprev) + np.dot(B, U)
        k1 = h * f
        k2 = h * (f + k1 / 2)
        k3 = h * (f + k2 / 2)
        k4 = h * (f + k3)
        Xnext = Xprev + (k1 + 2 * k2 + 2 * k3 + k4) / 6
        X = np.c_[X, Xnext]
        Ytemp = np.dot(C, Xnext)
        Y = np.append(Y, Ytemp)
        T = np.append(T, t)
        if Udlt:
            U = 0
            Udlt = False
    return Y, T, X



import numpy as np
import matplotlib.pyplot as plt
import time_response as tresp
from math import log


# TODO:
# -root locus func
# -nyquist plots
# -function for model recognition
# -ss2tf
# -controlability and observability
# -L and K matrices
# -discrete time variants

# denominator and numerator coefficients must be given in a specific order:
# numerator: a1*s^n + a2*s^(n-1) + a3*s^(n-2) + a4*s^(n-3) + ... + an =>  [a1, a2, a3, a4, ..., an]
# denominator: b1*s^n + b2*s^(n-1) + b3*s^(n-2) + b4*s^(n-3) + ... + bn => [b1, b2, b3, ..., bn]


class sys:
    def __init__(self):
        self.A = np.array([])
        self.B = np.array([])
        self.C = np.array([])
        self.D = np.array([])
        self.TFnumerator = np.array([])
        self.TFdenominator = np.array([])
        self.startCond = np.array([])

    # diagonal form
    def diag(self):  # TODO fix this
        eigVals, eigVecs = np.linalg.eig(self.A)
        P = np.empty((np.size(eigVals), 0))  # transformation matrix
        for eigenVector in eigVecs:
            P = np.column_stack((P, eigenVector))
        invP = np.linalg.inv(P)
        A = np.matmul(invP, np.matmul(self.A, P))
        B = np.matmul(invP, self.B)
        C = np.matmul(self.C, P)
        return A, B, C

    # observable canonical form of SISO system
    # highest power of the s in denominator must be bigger than highest power in the numerator
    # highest power of s in denominator must have coefficient equal 1
    def obsv(self):
        num = self.TFnumerator
        den = self.TFdenominator
        numCols = np.size(self.TFnumerator)
        denCols = np.size(self.TFdenominator)
        if denCols <= numCols:
            raise Exception('denominator must be higher order than numerator')
        if den[0] != 1:
            div = den[0]
            num = num / div
            den = den / div
        den = den[1:]
        numCols = np.size(num)
        denCols = np.size(den)
        if denCols > numCols:  # making numerator and denominator equal length filling with 0
            tempArr = np.zeros(denCols)
            tempArr[denCols - numCols:] = num
            num = tempArr
        obsvA = np.array([-1 * den]).transpose()  # column of minus denumerator values
        subMat = np.identity(denCols - 1)  # identity matrix
        subMat = np.vstack((subMat, np.zeros(denCols - 1)))  # row of 0 added to identity matrix
        obsvA = np.append(obsvA, subMat, axis=1)
        obsvB = num[..., None]  # column of numerator values
        obsvC = np.array([1])
        obsvC = np.append(obsvC, np.zeros(denCols - 1), axis=0)  # row of 0
        return obsvA, obsvB, obsvC

    # controllable canonical form of SISO system
    # highest power of the s in denominator must be bigger than highest power in the numerator
    # highest power of s in denominator must have coefficient equal 1
    def contr(self):
        num = self.TFnumerator
        den = self.TFdenominator
        numCols = np.size(self.TFnumerator)
        denCols = np.size(self.TFdenominator)
        if numCols >= denCols:
            raise Exception('denominator must be higher order than numerator')
        if den[0] != 1:
            div = den[0]
            num = num / div
            den = den / div
        den = den[1:]
        numCols = np.size(num)
        denCols = np.size(den)
        if denCols > numCols:  # making numerator and denominator equal length filling with 0
            tempArr = np.zeros(denCols)
            tempArr[denCols - numCols:] = num
            num = tempArr
        contA = np.zeros(denCols - 1)[..., None]  # column of 0
        subMat = np.identity(denCols - 1)  # identity matrix
        contA = np.append(contA, subMat, axis=1)
        col = np.array([-1 * den])
        contA = np.append(contA, col, axis=0)
        contB = np.zeros(denCols - 1)[..., None]  # column of 0
        contB = np.vstack((contB, np.array([1])))
        contC = np.array(num)  # row of numerator values
        return contA, contB, contC


class ss(sys):
    def __init__(self, A, B, C, D=[0], stCond=None):
        super().__init__()
        if stCond is None:
            stCond = []
        self.A = np.array(A)
        tempB = np.array(B)
        self.B = np.reshape(tempB, (np.shape(tempB)[0], 1))
        self.C = np.array(C)
        self.D = np.array(D)
        if np.size(np.array(stCond)) < np.size(A, axis=0):  # if starting conditions vector is too short
            self.startCond = np.array(stCond)
            for i in range(np.size(A, axis=0) - np.size(self.startCond)):
                np.append(self.startCond, np.array([0]))
        else:
            raise Exception('number of starting conditions must be equal to the order of the system')

    def __str__(self):
        strA = str(self.A)
        strB = str(self.B)
        strC = str(self.C)
        strD = str(self.D)
        strRes = 'A = ' + strA + '\n' + 'B = ' + strB + '\n' + 'C = ' + strC + '\n' + 'D = ' + strD + '\n'
        return strRes


class tf(sys):
    def __init__(self, num, denum):
        super().__init__()
        self.TFnumerator = np.array(num)
        self.TFdenominator = np.array(denum)

    def __mul__(self, otherSys):
        newNumerator = np.polymul(self.TFnumerator, otherSys.TFnumerator)
        newDenominator = np.polymul(self.TFdenominator, otherSys.TFdenominator)
        newSys = tf(newNumerator, newDenominator)
        return newSys

    def __str__(self):
        numerator = ''
        denominator = ''
        line = ''
        strRes = ''
        if np.size(self.TFnumerator) == 1:
            numerator += str(self.TFnumerator[0])
        else:
            for n, num in enumerate(self.TFnumerator):
                if num != 0:
                    if n == np.size(self.TFnumerator) - 1:
                        numerator += ' + ' + str(num)
                    else:
                        if n == 0:
                            pass
                        else:
                            numerator += ' + '
                        numerator += str(num) + f's^{np.size(self.TFnumerator) - n - 1}'
        for n, num in enumerate(self.TFdenominator):
            if num != 0:
                if n == 0:
                    denominator += f'{num}s^{np.size(self.TFdenominator) - 1}'
                else:
                    if n == np.size(self.TFdenominator) - 1:
                        denominator += ' + ' + str(num)
                    else:
                        denominator += ' + ' + str(num) + f's^{np.size(self.TFdenominator) - n - 1}'

        largestLength = max(len(numerator), len(denominator))
        for i in range(largestLength):
            line += '-'
        line += '--\n'
        strRes += numerator + '\n' + line + denominator + '\n'
        return strRes


# returns ss system in a observable canonical form based on given tf
def tf2ss(system):
    if isinstance(system, tf):
        A, B, C = system.obsv()
        D = [0]
        systemSS = ss(A, B, C, D)
        return systemSS
    else:
        raise Exception('you need to pass tf system as the argument')


def ss2tf(system):
    if isinstance(system, ss):
        pass
    else:
        raise Exception('you need to pass ss system as the argument')


# returns array of poles of system
def poles(system):
    roots = np.array([])
    if isinstance(system, tf):
        roots = np.roots(system.TFdenominator)
    elif isinstance(system, ss):
        roots, eigvecs = np.linalg.eig(system.A)
    else:
        raise Exception('argument must be tf or ss system')
    return roots


# returns array of zeros of system
def zeros(system):
    if isinstance(system, tf):
        roots = np.roots(system.TFnumerator)
    elif isinstance(system, ss):
        roots, eigvecs = np.linalg.eig(system.A)
    else:
        raise Exception('argument must be tf or ss system')
    return roots


# draws phase portrait on phase plane of min 2nd order system, var1 and var2 describes which state variables need to
# be plot: 0 - x1 1 - x2 etc.
def phasePortrait(system, Xinit=None, var1=0, var2=1, t=1):
    if Xinit is None:
        Xinit = [[-1, -2], [-3, -4], [5, 6], [7, 8]]
    if isinstance(system, tf):
        systemSS = tf2ss(system)
        phasePortrait(systemSS, var1, var2)
    elif isinstance(system, ss):
        fig = plt.figure()
        ax = fig.add_subplot()
        plt.grid(linestyle='--')
        ax.set_title('Phase portrait')
        ax.set_ylabel(f'x{var1 + 1}(t)')
        ax.set_xlabel(f'x{var2 + 1}(t)')
        legendList = []
        if np.shape(system.A)[0] >= 2:
            Xinit = np.array(Xinit)
            for i in range(np.shape(Xinit)[0]):
                Y = T = Xtmp = X = np.array([[]])
                Y, T, Xtmp = tresp.solveRK4(system, 0, Xinit[i], time=t)
                X = Xtmp
                stateNum = np.shape(system.A)[0]
                XrowsNum = np.shape(X)[0]
                if stateNum >= var1 or stateNum >= var2:
                    # for k in range(0, XrowsNum, stateNum):
                    #     ax.plot(X[k + var1], X[k + var2])
                    ax.plot(X[var1], X[var2])
                    legendList.append(f'xo={Xinit[i]}')
                else:
                    raise Exception("system doesn't have that many state variables. Lower var1 or var2")

            plt.legend(legendList)
            plt.show()
        else:
            raise Exception('system needs to be at least 2nd order')


# takes system and returns tuple of vectors (Y,T,X) - response and time vectors and state trajectory matrix
def step(system, Tpts=None, plot=False, solver='rk4'):
    Y = T = X = np.array([])
    if isinstance(system, tf):
        systemSS = tf2ss(system)
        step(systemSS)
    elif isinstance(system, ss):
        if solver == 'ee':  # explicit (forward) Euler
            Y, T, X = tresp.solveEE(system, 1)
        elif solver == 'ie':  # implicit (backward) Euler
            Y, T, X = tresp.solveIE(system, 1)
        elif solver == 'trap':  # trapezoidal
            Y, T, X = tresp.solveTrap(system, 1)
        elif solver == 'rk4':
            Y, T, X = tresp.solveRK4(system, 1)
        else:
            raise Exception('wrong solver chosen, choose: ee, ie, trap or rk4')
        if Tpts is not None:
            Yi = np.interp(Tpts, T, Y)
            Y = Yi
            T = Tpts
        if plot:
            fig = plt.figure()
            ax = fig.add_subplot()
            ax.plot(T, Y)
            ax.set_title('Step response')
            ax.set_ylabel('y(t)')
            ax.set_xlabel('t [s]')
            plt.show()
        else:
            return Y, T, X
    else:
        raise Exception('argument must be a tf or ss system')


# takes system and returns tuple of vectors (Y,T) - response and time vectors
def pulse(system, plot=False, solver='rk4'):
    Y = T = X = np.array([])
    if isinstance(system, tf):
        systemSS = tf2ss(system)
        step(systemSS)
    elif isinstance(system, ss):
        if solver == 'ee':  # explicit (forward) Euler
            Y, T, X = tresp.solveEE(system, 'delta')
        elif solver == 'ie':  # implicit (backward) Euler
            Y, T, X = tresp.solveIE(system, 'delta')
        elif solver == 'trap':  # trapezoidal
            Y, T, X = tresp.solveTrap(system, 'delta')
        elif solver == 'rk4':
            Y, T, X = tresp.solveRK4(system, 'delta')
        else:
            raise Exception('wrong solver chosen, choose: ee, ie, trap or rk4')
        if plot:
            fig = plt.figure()
            ax = fig.add_subplot()
            ax.plot(T, Y)
            ax.set_title('Impulse response')
            ax.set_ylabel('y(t)')
            ax.set_xlabel('t [s]')
            plt.show()
        else:
            return Y, T, X
    else:
        raise Exception('argument must be a tf or ss system')


# python's pow() function couldn't handle complex numbers and was trying to cast it into something else
def __imagPow(base, power):
    res = 1
    for i in range(power - 1):
        res *= base
    return res


# creates sinusodial transfer function G(jw)
def __sinTF(system):
    num = []
    den = [1]
    if isinstance(system, tf):
        for i in range(len(system.TFnumerator)):
            num.append(complex(system.TFnumerator[i], 0))
        for i in range(len(system.TFdenominator)):
            den.append(complex(system.TFdenominator[i], 0))
        for i in range(len(num)):
            num[i] = num[i] * __imagPow(1j, len(num) - i)
        for k in range(len(den)):
            den[k] = den[k] * __imagPow(1j, len(den) - k)
        return num, den
    elif isinstance(system, ss):
        systemTF = ss2tf(system)
        __sinTF(systemTF)


# draws bode diagrams for system
def bode(system, plot=False):
    n, d = __sinTF(system)
    num = den = 0
    Gvec = np.array([])
    Phvec = np.array([])
    Wvec = np.array([])
    # calculating the amplitude and phase characteristics
    for w in np.linspace(0.1, 1000, 20000):
        for i in range(len(n)):
            num += n[i] * pow(w, len(n) - i)
        for k in range(len(d)):
            den += d[k] * pow(w, len(d) - k)
        G = 20 * log(abs(num / den))
        Ph = np.angle((num / den), deg=True)
        Wvec = np.append(Wvec, w)
        Gvec = np.append(Gvec, G)
        Phvec = np.append(Phvec, Ph)
    if plot:
        fig, ax = plt.subplots(2)
        plt.grid(linestyle='--')
        ax[0].set_title('Bode diagrams')
        ax[0].set_ylabel('Magnitude [dB]')
        ax[0].set_xlabel('ω [rad/s]')
        ax[0].set_xscale('log')
        ax[1].set_ylabel('Phase shift [°]')
        ax[1].set_xlabel('ω [rad/s]')
        ax[1].set_xscale('log')
        ax[0].plot(Wvec, Gvec)
        ax[1].plot(Wvec, Phvec)
        plt.show()
    else:
        return Gvec, Phvec, Wvec


# draws nyquist plot of system
# FIXME weird plots (a bit different to matlab/octave)
def nyquist(system):
    n, d = __sinTF(system)
    num = den = 0
    Pvec = np.array([])
    Qvec = np.array([])
    Wvec = np.array([])
    for w in np.linspace(0.1, 1000, 40000):
        for i in range(len(n)):
            num += n[i] * pow(w, len(n) - i)
        for k in range(len(d)):
            den += d[k] * pow(w, len(d) - k)
        P = abs(num / den)
        Q = (num / den).imag
        Wvec = np.append(Wvec, w)
        Pvec = np.append(Pvec, P)
        Qvec = np.append(Qvec, Q)
    fig = plt.figure()
    plt.grid(linestyle='--')
    ax = fig.add_subplot()
    ax.set_title('Nyquist plot G(jω) = P(ω) + jQ(ω)')
    ax.set_ylabel('Q(ω)')
    ax.set_xlabel('P(ω)')
    ax.plot(Pvec, Qvec)
    ax.plot(Pvec, -1 * Qvec)
    plt.show()


def rlocus(system, CLtf=tf([1], [1])):
    if isinstance(system, tf):
        k = 0
        OLpoles = np.roots(system.TFdenominator)
        OLzeros = np.roots(system.TFnumerator)
        temp1 = np.polynomial.polynomial.polymul(CLtf.TFdenominator, system.TFdenominator)
        temp2 = np.polynomial.polynomial.polymul(CLtf.TFnumerator, system.TFnumerator)
        CLdenominator = np.polynomial.polynomial.polyadd(temp1, temp2)
        CLnumerator = np.polynomial.polynomial.polymul(CLtf.TFdenominator, system.TFnumerator)
        CLpoles = np.roots(CLdenominator)
        CLzeros = np.roots(CLnumerator)
    # continue

    elif isinstance(system, ss):
        pass
    else:
        pass
